local json = require 'cjson'
local yaml = require 'yaml'

--

local defaultmeta = {
	producer = 'Paged Markdown 3'
}

local metamap = {
	author = '/Creator',
	producer = '/Producer',
	title = '/Title',
}

--

local pdfmeta = {}

local function pmdpnums2pdfpnums(pagenumbers)
	local res = {}
	local stylestm
	local labeledNum = 1

	local pnumstyle_map = {
		none = false,
		decimal = '/D',
		roman = '/R'
	}

	for i, p in ipairs(pagenumbers) do
		if p[2] or p[1] ~= stylestm then
			local def = {}

			if p[2] then labeledNum = 1 end
			if pnumstyle_map[p[1]] then def['/S'] = pnumstyle_map[p[1]] end
			if labeledNum ~= 1 then def['/St'] = labeledNum end

			table.insert(res, i - 1);
			table.insert(res, def);
			stylestm = p[1];
		end
		labeledNum = labeledNum + 1;
	end
	return res
end

function pdfmeta.annotate(filename_pdf, meta, pagenumbers)
	local qpdf = assert(io.popen(('qpdf --json-output "%s" -'):format(filename_pdf), 'r'))
	local doc = json.decode(qpdf:read '*a')
	qpdf:close()
	--

	for k, v in pairs(defaultmeta) do
		local overwritten = false
		for o, _ in pairs(meta) do
			if k == o:lower() then
				overwritten = true
				break
			end
		end
		if not overwritten then
			meta[k] = v
		end
	end

	for _, node in pairs(doc.qpdf[2]) do
		if type(node.value) == 'table' and node.value['/Creator'] then
			node.value['/Creator'] = nil -- clear default value from marionette
			for k, v in pairs(meta) do
				if metamap[k:lower()] then
					node.value[metamap[k:lower()]] = 'u:' .. v
				end
			end
		end
		if type(node.value) == 'table' and node.value['/Type'] == '/Catalog' then
			node.value['/PageLabels'] = {
				['/Nums'] = pmdpnums2pdfpnums(pagenumbers)
			}
		end
	end

	--
	json.encode_escape_forward_slash(false)
	local docmod = json.encode(doc)

	local tmpjson = os.tmpname()
	local out = assert(io.open(tmpjson, "w"))
	out:write(docmod)
	out:close()

	qpdf = assert(io.popen(('qpdf --json-input "%s" "%s"'):format(tmpjson, filename_pdf), 'r'))
	qpdf:close()
	os.remove(tmpjson)
end

return pdfmeta
